#!/usr/bin/perl -w
# .mdl to Binary Adjacency Matrix Utility
#
# Rogelio Oliva <roliva@tamu.edu>
# Mays Business School
# 
# Jun. 11, 2015. v3.0
# 
# History
#
# v0.1 01/1997. Core functionality for rLab
# v0.2 02/1997. Output format changed to match MATLAB
# v0.3 02/1998. Changed to work with Perl 5.1
# v0.4 02/1998. Implemented workaround Vensim 3.0 bug on document tool
# v0.5 02/1998. Implemented the creation of a vector w/variable names
# v1.0 06/2003. Implemented to work with *.mdl format 
#				Introduced the ability to processes models with arrays
#				Integrated MATLAB output to structured array
#				Fully automated generation of MATLAB code
#				Changed to work in Perl 5.6
#				Developed OSX functionality and application
# v1.1 11/2004. Fixed processing of variables with digits in name
#				Expanded name string to 60 characters
# v2.0 02/2006. Implement functionality over the Internet
# v3.0 06/2015. Reverted functionality to work as Perl script
#				Tested with Perl 5.16
#
#  Rogelio Oliva, 1997, 1998, 2003, 2004, 2006, 2015.
#
###############################################
# This utility creates the binary adjacency matrix and a column vector
# with all the variable names for a Vensim model (*.mdl format). 
# The output is an *.m file formatted to work with Matlab, specifically, 
# with the Model Structural Analysis (MSA) functions as described in Oliva (2004)
# http://dx.doi.org/10.1002/sdr.298. The MSA functions and documentation
# are available at http://iops.tamu.edu/faculty/roliva/research/sd/.
#
# Input file:
#	- Vensim *.mdl format file
# Output file:
#	- MATLAB *.m file with a function, named as the original model,
#	  that returns the adjacency matrix and the name list for model 
#	  variables into a single structured array with components .adj 
#	  and .nms.
#
# The following transformations are done to the model variables:
#	- variables introduced by Vensim (INITIAL TIME, etc.) are eliminated.
#	- variable names are cliped to 60 characters.
#	- apostrophes are removed from variable names 
#		-- difficult string caracter to handle in MATLAB.
#	- macros are treated as functions not as variables.
#	- arrays are eliminated
#		-- only the first equation for a variable is considered.
#	- links from stock-initialization equations are removed
#		-- not accurate if the initialization eq. contains a comma

$date = scalar localtime;
$version = "3.0";
print "\nVensim *.mdl file to Binary Matrix Utility v".$version."\n";
print "(c) Rogelio Oliva, 1997, 1998, 2003, 2004, 2006, 2015.\n";
print $date."\n\n";

$old_input_record_separator = $/;
undef $/;						 					#Change $/ to read until EOF


# Take the first argument from the command line as input 
$fileIN=shift @ARGV;
open(VENMDL, $fileIN) 
	or die "Could not open ".$fileIN.": $!\n";		#Open input file
$equation_list = <VENMDL>;
close VENMDL;										#Close file as soon as possible
print "   Processing ".$fileIN."\n";

#Generate the default output filename & create file
$fileOut = $fileIN;
$fileOut =~ s/\.\w*$//;								#Remove .ending, if any   
$fileM = $fileOut.".m";								# add .m
open(MFILE,">".$fileM) 								#Open output file
	or die "Could not open ".$fileOut.": $!\n";				
print "   Output in  ".$fileM."\n\n";

#the default file name will now be piped from the webwrapper as a second argument
#$fileOrig=shift @ARGV;

#Information for M file headings
$function_name = $fileIN;						#MATLAB function has the same name as file
$function_name =~ s/.*\///;						#Remove directory hierarchy (everything up to the last /)
$function_name =~ s/\.mdl//;						#Remove .m termination


#Write Matlab headings and documentations in M file
print MFILE "function a = ".$function_name."()\n";
print MFILE "%".$function_name."     Y=".$function_name."\n";
print MFILE "%     Returns to Y the adjacency matrix and the name list for model\n";
print MFILE "%     ".$fileIN."\n";
print MFILE "%     Y is a structured array with subcomponents:\n";
print MFILE "%         .adj -- sparse binary matrix with adjacency relationships\n";
print MFILE "%         .nms -- vector with variable names\n";
print MFILE "%\n";
print MFILE "%     Generated on ".$date.", with mdl2bin Utility v".$version.".\n";
print MFILE "%     (c) Rogelio Oliva, 2015.\n";
print MFILE "\n\n";
   
#Major transformations to the file as a whole
$equation_list =~ s/\\\n/\n/g; 												#get rid of vensim's internal escaped newlines.
$equation_list = &unspacify($equation_list); 								#get rid of extra spaces.
while ($equation_list =~ s/(:macro:.+?:end of macro:)//gsi) {}; 			#delete any macros, they are not processed by this script.
while ($equation_list =~ s/\s*\*{56}\s*\.[^\n]+?\s*\*{56}\~[^\|]*\|//gs) {};	#delete sector headings.
if ($equation_list =~ s/^(.+?)(\\\\\\---\/\/\/.*)$/$1/si) {} 				#delete the sketch too.
$equation_list=~ s/\{[^\}]+\}//g; 											#get rid of all curly brace comments (curly braces are not allowed within curly braces!)
$equation_list=~ s/'//g;													#eliminate all apostrophes

#delete Vensim introduced variables
$equation_list =~ s/INITIAL TIME  =[^\|]+\|//; 
$equation_list =~ s/FINAL TIME  =[^\|]+\|//;
$equation_list =~ s/SAVEPER  =[^\|]+\|//;
$equation_list =~ s/TIME STEP  =[^\|]+\|//;



$/ = $old_input_record_separator;	# back to line-per-line analysis
$i=0;								# initialize equation counter
$old_name="";						# variable to control for arrayed variables

# This is the main equation extraction rutine 
while ($equation_list=~/(.+?(?:\||$))/gs) {
	$eq_match=$1;
	next if ($eq_match =~ /^\s*$/);			# skip if eq is blank
	next if ($eq_match =~ / : /);			# skip if eq is an array index definition
	($canonical_name, $equation)=&split_and_clean_equation($eq_match);
	if ($canonical_name ne $old_name) {		# if variable has not been introduced (array eqs. are together in .mdl)
		$names[$i]=$canonical_name;			# add variable to variable name
		$equations[$i]=$equation;			# add equation to equation list
		$old_name=$canonical_name;			# update $oldname
		$i++;								# increase var/eq counter
	}
}



# Print out the names vector (40 characters in with)
# MFile subindex is incremented by one because 
# Perl arrays run from 0 while MATLAB arrays run from 1
# Explicit subindex to control the equation number
for ($j=0; $j<$i; $j++) {
	$m_name=substr($names[$j],0,60);
	print MFILE "names(".($j+1).",:)='".$m_name." " x(60-length($m_name))."';\n";
}
print MFILE "a.nms=names;\n\n";		#MATLAB code to pass to output variable


# Develop hash for variable names
# Index in $name_index is already incremented by one
my %name_index;
for ($j=0; $j<$i; $j++) {$name_index{$names[$j]}=$j+1;}


# Develop hash for fucntions with initialization values
# "DELAY N" "DELAY MATERIAL" and "DELAY PROFILE" not implemente yet
# because initialization value is not the last input to the function
@functions_w_inits=("INTEG",
					"SMOOTHI","SMOOTH3I",
					"DELAY1I","DELAY3I",
					"DELAYED FIXED","DELAYED INFORMATION");
my %functions_w_inits;
for (@functions_w_inits) {$functions_w_inits{$_}++;}


# Extract and print adjacency matrix based on predecesors
print MFILE "adj=sparse(zeros(".$i.",".$i."));\n";
for ($j=0; $j<$i; $j++) {											# for every equation
	$equations[$j] =~ s/\[[^\]]*\]//g;								# remove subindeces
	@elements=&extract_elements($equations[$j]);					# extract elements
	if ($elements[1]) {												# if not a constant
		if ($functions_w_inits{$elements[1]}) {						# if equation is a function_w_initials
			$equations[$j] =~ s/\,[^\,]*\)$//;						# truncate from last comma to the last paren
			@elements=&extract_elements($equations[$j]);			# re-do extraction
		}
## this is extra stuff that I had to introduce to diagnose
#		print MFILE $equations[$j]."\n";
#		for ($k=1; $k<$t; $k++) {
#			print MFILE $elements[$k]."...";
#		}
#		print MFILE "\n";
## end of extra stuff to diagnose

		for ($k=1;$k<$t; $k++) {									# for every word in the rhs (subindex>0)
			if ($name_index{$elements[$k]}) {	# only if already exist as a variable in the lhs -- this eliminates all of Vensim special words
				print MFILE "adj(".($name_index{$elements[$k]}).",".($j+1).")=1;\n";	#note that j is accessing equations (from 0) 
			}																			#while $name_index has already been shifted.
		}
	}
}
print MFILE "a.adj=adj;\n";		#MATLAB code to pass to output variable
close (MFILE);

#(\(.*,.*\))*

#------------------------------
# the following while takes out all the individual elements of the equation
# word[0] contains the dependent variable, constants and blanks are eliminated
sub extract_elements {
	my ($eq)=shift;
	my (@words)=();
	$t=0;
	while ($eq =~ /([^=\*\[\],\/+\-\(\)\^><:]+(?:\[[^\]]+\]\s*)?)([=\*\/+\-\(\),\^><:]*)/g) {
		$words[$t]=&canonicalize(&unspacify2($1));
		if ($words[$t]=~/[a-df-z]+/i) {		# if word is not a constant (has at least one not 'E' letter),
			$t++;							# ingrement t - otherwise, it is replaced
		}	 
	}
	return(@words);
}

sub split_and_clean_equation {
	my ($arg)=shift;
	my($real_name, $canonical_name);
	my ($equation,$unit,$def)=split(/~/,$arg,3);
	$unit=~s/\[[^\]]*\]//g;
	$equation=&superunspacify($equation);
	$unit=&superunspacify($unit);
	$def=&superunspacify($def);
	if ($equation =~ /^([^:=(\n]+)((?:(?::[A-Za-z]+:|:=|=)\s*)*)(.*)$/) {
		$real_name = &unspacify2($1);
		$canonical_name = &canonicalize($real_name);
	} else {
		print "Equation didn't match.\nEquation: \"$arg\"\nMatch didn't work in sub: \"split_and_clean_equation\" \n";
	}
	return($canonical_name,$equation);	# Note that the rutine could also export $real_name $unit and $def
}

sub canonicalize {
	my ($var) = shift;
	return ( &unspacify(&remove_subscripts($var)) );
}

sub remove_subscripts {
	my($a) = shift;
	$a=~s/\[[^\]]+\]//;
	return($a);
}

sub superunspacify {
	my($a)=shift;
	$a=~s/[\s\n\\]+/ /g;
	$a=&unspacify($a);
	$a=~s/ +\[/\[/g;
	$a=~s/\[ +/\[/g;
	$a=~s/ +\]/\]/g;
	return($a);
}

sub unspacify2 {
	my($a)=shift;
	$a=&unspacify($a);
	$a=~s/ *, */,/;
	$a=~s/ *! *\]/!\]/;
	$a=~s/^if then else$/IF THEN ELSE/i;
	$a=~s/^power$/POWER/i;
	return($a);
}

#gets rid of leading and trailing space
sub unspacify {
	my($a)=shift;
	$a=~s/\s*$//;
	$a=~s/^\s*//;
	return($a);
}


